#include <stdio.h>


#include "map.h"
#include <vcg/complex/trimesh/update/bounding.h>
#include <vcg/complex/trimesh/update/normal.h>
#include <vcg/complex/trimesh/allocate.h>

#include <GL\gl.h>

#include <wrap/gl/space.h>


Map map;

const int RIVER=8;
void Map::UpdateRivers(){
  for (VertexIterator  vi=vert.begin(); vi!=vert.end(); vi++){
    vi->nriver=0;
  }
  for (FaceIterator   fi=face.begin(); fi!=face.end(); fi++){
    if (fi->zoneid==RIVER) {
      fi->V(0)->nriver++;
      fi->V(1)->nriver++;
      fi->V(2)->nriver++;
    }
  } 
}

void Map::ShrinkRivers(){
  UpdateRivers();
  
  vcg::Point3f tmp[vert.size()];
  int i=0;
  VertexIterator  vi;
  FaceIterator  fi;
  
  for (vi=vert.begin(),i=0; vi!=vert.end(); vi++,i++){
    tmp[i]=vi->P();
  }
  const float K=0.05;
  
  for (fi=face.begin(); fi!=face.end(); fi++){
    for (int h=0; h<2; h++){
      int k=(h+1)%3;
      if ((fi->V(h)->nriver>1) && (fi->V(k)->nriver>1)){
        vcg::Point3f d=(fi->V(h)->P()-fi->V(k)->P());
        int i,j;
        i=fi->V(h)-&(vert[0]);
        j=fi->V(k)-&(vert[0]);
        tmp[i]-=d*K;
        tmp[j]+=d*K;
      }
    }
  } 
 
  for (vi=vert.begin(),i=0; vi!=vert.end(); vi++,i++){
    vi->P()=tmp[i];
  }
}

void Map::ComputeShade(){
  for (FaceIterator fi=face.begin(); fi!=face.end(); fi++) fi->ComputeShade();
}

void CFace::ComputeShade(){
//  shade=vcg::NormalizedNormal< CFace >(*this).Z();
  static vcg::Point3f L=vcg::Point3f(0,-1,3).Normalize();
  shade=vcg::NormalizedNormal< CFace >(*this)*L;
}

inline float cross2(vcg::Point3f a, vcg::Point3f b){
  return (a[0]*b[1]-a[1]*b[0]);
}

vcg::Point3f CFace::GetN() const{
  return vcg::NormalizedNormal< CFace >(*this);
}

bool CFace::GetBariCoords(vcg::Point3f p, float &a, float &b, float &c ) const
{	

  
	float a0=cross2( (P(2) - P(1)) , (p - P(1)) );
	float a1=cross2( (P(0) - P(2)) , (p - P(2)) );
	float a2=cross2( (P(1) - P(0)) , (p - P(0)) );

  if (((a0>=0) != (a1>=0) ) || ((a0>=0) != (a2>=0) ) ) return false;
  
  a=a0/(a0+a1+a2);
  b=a1/(a0+a1+a2);
  c=a2/(a0+a1+a2);
  
	return true;
}

static Map::ConstFaceIterator lastfound;

float DistFlat(vcg::Point3f d){return d[0]*d[0]+d[1]*d[1];};

vcg::Point3f  Map::GetClosestVertP(vcg::Point3f p) const{
  vcg::Point3f res; float mind=999999.0;
  for (ConstVertexIterator v=vert.begin(); v!=vert.end(); v++) {
    float dv=DistFlat(v->P()-p);
    if (dv<mind) { mind=dv; res=v->P(); }
  }
  return res;
}

float Map::GetZ(vcg::Point3f p) const{
  lastfound=face.end();
  for (ConstFaceIterator fi=face.begin(); fi!=face.end(); fi++) {
    float a,b,c;
    if (fi->GetBariCoords(p,a,b,c)) {
    lastfound=fi;
    return fi->P(0).Z()*a+
           fi->P(1).Z()*b+
           fi->P(2).Z()*c;
    }
  }
  return 0; // not found
}

vcg::Point3f Map::GetN() const {
  if (lastfound!=face.end()) return lastfound->GetN();
}

bool Map::Load(char *fn)
{
  Clear();
  
  FILE *f=fopen(fn,"rt");
  if (!f) {
    printf("Cannot open map \"%s\"\nwrong path? check MapScribbler.cfg\n",fn);
    return false;
  }
  
  int nv;
  fscanf(f,"%d",&nv);

  VertexIterator   vi = vcg::tri::Allocator<Map>::AddVertices(*this,nv);
  for (int i=0; i<nv; i++,vi++){
    fscanf(f,"%f %f %f", &(vi->P().X()),&(vi->P().Y()),&(vi->P().Z()));
    
    //vi->P().Z()*=-1;
  }

  int nf;
  fscanf(f,"%d",&nf);
  
  FaceIterator   fi = vcg::tri::Allocator<Map>::AddFaces(*this,nf);
  for (int i=0; i<nf; i++, fi++){
    int a,b,c,d,e,ff;
    fscanf(f,"%d %d %d %d %d %d",&d,&e,&ff,&a,&b,&c);
    fi->zoneid=d;
    fi->V(0)=&(vert[a]);
    fi->V(1)=&(vert[b]);
    fi->V(2)=&(vert[c]);
  }
  
  fclose(f);
  vcg::tri::UpdateBounding<Map>::Box(*this);
  ComputeShade();
  return true;
}

class Color3f{
public:
  unsigned char rgb[3];
  void SendGl(){
    glColor3ubv(rgb);
  }
  void SendGl(float dark){
    float h=dark/255;
    glColor3f(rgb[0]*h,rgb[1]*h,rgb[2]*h);
  }
  void SendGlDark(float dark){
    float h=dark/280;
    glColor3f(rgb[0]*h,rgb[1]*h,rgb[2]*h);
  }
  void SendGlDark(){
    glColor3d(rgb[0]/280.0,rgb[1]/280.0,rgb[2]/280.0);
  }
  
};

std::vector<vcg::Point2f> locations; // cityes etc

char* ReadLine(FILE *a);

bool Map::LoadLocations(char* filename){
  FILE* f=fopen(filename,"rt");
  if (!f) return false;
  
  int orig=0;
  char* tmpst;
  ReadLine(f);
  tmpst=ReadLine(f);  
  sscanf(tmpst,"%d", &orig);
  for (int i=0; i<orig; i++) {
    tmpst=ReadLine(f);
    int a[12];
    unsigned int code=0;
    float x=100,y=100;
    char sta[256],stb[256];
    sscanf(tmpst, " %d %d %s %s %u %d %d %d %d %d %d %d %d %f %f",
               a,a+1,sta,stb,&code,a+3,a+4,a+5,a+6,a+7,a+8,a+9,a+10,&x, &y );
    //printf("Location at %f %f (%u) \"%s\"\n",x,y,code,tmpst);
    if (//(x>-99) && (x<99) && (y>-99) && (y<99) 
        (x!=-100)
       && ((code>>24)>1)){
      location.push_back(vcg::Point2f(x,y));
    }
    ReadLine(f);
  }
  fclose(f);
  return true;
}

const int MAXZONES=256;
Color3f mapColor[MAXZONES];

bool Map::ReloadMapColors(){
  FILE *f=fopen("mapcolors.txt","rt");
  if (!f) return false;
  int n;
  fscanf(f,"%d",&n);
  for (int i=0; i<n; i++){
    int a,b,c,d;
    fscanf(f,"%d %d %d %d",&a,&b,&c,&d);
    mapColor[a%MAXZONES].rgb[0]=b;
    mapColor[a%MAXZONES].rgb[1]=c;
    mapColor[a%MAXZONES].rgb[2]=d;
  }
  //for (int i=0; i<MAXZONES; i++){
  //  mapColor[i].rgb[0]=0;
  //  mapColor[i].rgb[1]=0;
  //  mapColor[i].rgb[2]=255;
  //}  
  fclose(f);
  return true;  
}

bool Map::SaveMapColors(){
  bool used[MAXZONES];
  for (int i=0; i<MAXZONES; i++) {
    used[i]=false;
  }
  
  int nused=0;
  for (FaceIterator   fi=face.begin(); fi!=face.end(); fi++){
    int z=(fi->zoneid)%MAXZONES;
    if (!used[z]) {
      used[z]=true;
      nused++;
    }
  }
  
  FILE *f=fopen("mapcolors.txt","wt");
  if (!f) return false;

  fprintf(f,"%d\n",nused);
  
  for (int i=0; i<MAXZONES; i++) {
    if (used[i]) {
      int col=((i+40)*2312)%256;
      fprintf(f,"%d %d %d %d\n",i,col,col,col);
    }
  }
  
  fclose(f);
  return true;  
}



bool Map::Save(char *fn)
{
  
  FILE *f=fopen(fn,"wt");
  if (!f) return false;
  
  
  int nv=vert.size();
  fprintf(f,"%d\n",nv);

  for (VertexIterator  vi=vert.begin(); vi!=vert.end(); vi++){
    fprintf(f,"%f %f %f \n", (vi->P().X()),(vi->P().Y()),(vi->P().Z()));
  }

  int nf=face.size();
  fprintf(f,"%d\n",nf);
  
  for (FaceIterator   fi=face.begin(); fi!=face.end(); fi++){
    fprintf(f,"  %d 0  3  %d  %d  %d \n",fi->zoneid,fi->V(0)-&(vert[0]),fi->V(1)-&(vert[0]),fi->V(2)-&(vert[0]));
  }
  fclose(f);
  return true;
}

void Map::Render(){
  glScalef(2.0/bbox.Diag(),2.0/bbox.Diag(),1);
  vcg::glTranslate(-bbox.Center());
  
  glBegin(GL_TRIANGLES);
  for (FaceIterator   fi=face.begin(); fi!=face.end(); fi++){
    mapColor[fi->zoneid].SendGl(fi->shade);
    for (int i=0; i<3; i++){
      glVertex2fv( &(fi->V(i)->P().X()) );
    }
  }
  glEnd();
  glColor3f(0,0,0);

  
  for (FaceIterator   fi=face.begin(); fi!=face.end(); fi++){
    glBegin(GL_LINE_LOOP);
    mapColor[fi->zoneid].SendGlDark(fi->shade);
    for (int i=0; i<3; i++){
      glVertex2fv( &(fi->V(i)->P().X()) );
    }
    glEnd();
  }
  
  if (0) {
  glBegin(GL_LINES);
  for (int x=(int)bbox.min.X(); x<bbox.max.X(); x++){
    glVertex2f(x,bbox.min.Y());
    glVertex2f(x,bbox.max.Y());
  }
  for (int y=(int)bbox.min.Y(); y<bbox.max.Y(); y++){
    glVertex2f(bbox.min.X(),y);
    glVertex2f(bbox.max.X(),y);
  }
  glEnd();
  }
  
  // draw cities
  glEnable(GL_BLEND);
  float s=1.5,z= -5.5;
  for (int i=0; i<location.size(); i++){
    glBegin(GL_TRIANGLE_FAN);      
    glColor4f(0.6,0.5,0.4,0.05);
    glVertex3d( location[i].X() , location[i].Y(), z);
    glColor4f(0.6,0.5,0.4,0.80);
    glVertex3d( location[i].X()+s, location[i].Y(), z);
    glVertex3d( location[i].X(), location[i].Y()+s, z);
    glVertex3d( location[i].X()-s, location[i].Y(), z);
    glVertex3d( location[i].X(), location[i].Y()-s, z);
    glVertex3d( location[i].X()+s, location[i].Y(), z);
    glEnd();
  }
  glDisable(GL_BLEND);


}
