
#include "texture.h"

#include <SDL/SDL.h>
#include <SDL/SDL_ttf.h>
#include <SDL/SDL_image.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <stdio.h>
#include <math.h>

#include "config.h"

static int cw[256], ch[256]; // letter height and width
static int cx[256], cy[256], // letter corner in texture space
           ct[256]; // letter max top
static bool present[256]; // is characte4r present

static int globalh;
int TSIZEX=0,TSIZEY=0;

typedef unsigned char byte;

static byte *buffer=NULL;
  
int Texture::nletters=0;

bool Texture::CanUseChar(int c){
  if (c<0  || c>255) return false;
  return present[c];
}

bool Texture::SaveAsDXT3(char *filename){
  FILE *f=fopen(filename,"wb");
  int nmipmaps=0;
  while ( 
    ((1<<nmipmaps) < TSIZEX) || 
    ((1<<nmipmaps) < TSIZEY) 
  ) nmipmaps++;
  
  nmipmaps++;
  
  unsigned int header[32] = {
  542327876,124,659463,1024, 1024,1024*1024,0,10,
  0,0,0,0,0,0,0,0,
  0,0,0,32, 4, 861165636,0,0,
  0,0,0,4198408,0,0,0,0,
  };
  header[3]=TSIZEY;
  header[4]=TSIZEX;
  header[5]=TSIZEY*TSIZEX;
  header[7]=nmipmaps;
  fwrite(header,4,32,f);
  int sx=TSIZEX,sy=TSIZEY; 
  byte *alphabuffer=(byte*) calloc(1,TSIZEX*TSIZEY);
  for (int i=0; i<TSIZEX*TSIZEY; i++) alphabuffer[i]=buffer[i*4+3];
  
  byte tmp[16]; 
  tmp[8]=tmp[9]=0xFF;
  tmp[10]=tmp[11]=tmp[12]=tmp[13]=tmp[14]=tmp[15]=0x00;
  
  int oldsx=sx;
  for (int mipmap=0; mipmap<nmipmaps; mipmap++) {
    for (int y=0; y<sy; y+=4)
    for (int x=0; x<sx; x+=4) {

      for (int dy=0,i=0; dy<4; dy++) 
      for (int dx=0; dx<4; dx++,i++)
      {
        int alpha=(int) (alphabuffer[ x+dx+(y+dy)*TSIZEX ]*transparencyLOD(mipmap));
        alpha>>=4;
        if (i&1) tmp[i/2] += alpha<<4;
        else tmp[i/2] = alpha;
      }
      fwrite(tmp,1,16,f);
    }

    sx/=2; sy/=2;
    if (sx<1) sx=1; if (sy<1) sy=1;
    for (int y=0; y<sy; y++)
    for (int x=0; x<sx; x++)
    {
      alphabuffer[x+y*oldsx]=(
        ((int)alphabuffer[x*2+  (y*2  )*oldsx]) +
        ((int)alphabuffer[x*2+1+(y*2  )*oldsx]) +
        ((int)alphabuffer[x*2  +(y*2+1)*oldsx]) +
        ((int)alphabuffer[x*2+1+(y*2+1)*oldsx]) +2)/4;
        
    }
  }
  fclose(f);
  free(alphabuffer);
  return true;
}

bool Texture::SaveAsPam(char *filename){
  FILE *f=fopen(filename,"wt");
  fprintf(f,"P7\nWIDTH %d\nHEIGHT %d\nDEPTH 4\nMAXVAL 255\nTUPLTYPE RGB_ALPHA\nENDHDR\n",
          TSIZEX,TSIZEY);
  //for (int i=0; i<TSIZEX*TSIZEY*4; i++) {
  //  fprintf(f,"%d ",(int)(buffer[i]));
  //}
  for (int y=0,i=3; y<TSIZEY; y++) {
  for (int x=0; x<TSIZEY; i+=4,x++) {
    fprintf(f,"%01X ",(((int)buffer[i]))>>4);
  }
  fprintf(f,"\n");
  }
  fclose(f);
  return true;
}

bool PNGSaveWithAlpha( const char * filename, const byte * const data, int sx, int sy, int reverse );

bool Texture::SaveAsPng(char *filename){
  return  false; //PNGSaveWithAlpha(filename, buffer, TSIZEX, TSIZEY,0);
}

void Texture::LoadGl(){
  gluBuild2DMipmaps(
    GL_TEXTURE_2D, 
    GL_RGBA, 
    TSIZEX, TSIZEY, 
    GL_RGBA,  
    GL_UNSIGNED_BYTE, 
    (void*)buffer
  );
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
}
  
bool Texture::Build(char* charmap, char *fontFilename, int sx, int sy){
  
  TSIZEX=sx; TSIZEY=sy;
  
  int MAXC=strlen(charmap);
  /*
  //const int \=(('z'-'a')+1)*2+1; 
  char charmap[MAXC];
  {
    int i=0;
    for (char c='a'; c<='z'; c++) charmap[i++]=c;
    for (char c='A'; c<='Z'; c++) charmap[i++]=c;
    charmap[i++]='-';
  }*/

  char st[2]; st[1]=0;
  
  
  SDL_Color font_col;
  font_col.r=255; 
  font_col.g=255;
  font_col.b=255; 

  static bool once=true;
  if (once) {
    if(TTF_Init()==-1) {
      printf("Unable to initialize font library\n");
      return false;
    }
  }  
  once=false;
  
  
  for (int i=0; i<256; i++) { ch[i]=-1; };

  // compute appox ideal size  
  nletters=0;
  for (int i=0; i<255; i++) present[ i ]=false;
  for (int i=0; i<MAXC; i++) present[ charmap[i] ]=true;
  for (int i=0; i<255; i++) if (present[ i ]) nletters++;
  if (nletters==0) nletters=1;

  int CSIZEW=(int)sqrt(TSIZEX*TSIZEY/(float)nletters)+5
     ,CSIZEH;
  
  if (buffer) free(buffer);
  buffer=(byte*)calloc(1,TSIZEX*TSIZEY*4);
  if (!buffer) printf("Out of memory??");
  
while (1) {
    static int times=0; if (times++>200) return false;
    
    static TTF_Font *font=NULL;
    if (font) TTF_CloseFont(font);
    font=TTF_OpenFont(fontFilename, CSIZEW);
    if (!font) {
      printf("Unable to load font at size %d\n",CSIZEW);
      return false;
    }
    CSIZEH=-TTF_FontDescent(font)+TTF_FontAscent(font)+1;
 
  // find a place for each letter  
  
  for (int i=0; i<256; i++) { present[i]=false; };

  int posx=2, posy=0;
  bool toobig=false;
  
  // clear texture
  for (int i=0; i<TSIZEX*TSIZEY*4; i+=4) {
    buffer[i]=buffer[i+1]=buffer[i+2]=255;
    buffer[i+3]=(DEBUG_MAPEDIT)?255:0;
  }
  
  for (int i=0; i<MAXC; i++) {
    char c=st[0]=charmap[i];
    
    if (present[c]) continue;
    present[c]=true;
    
    SDL_Surface *text_surface=TTF_RenderText_Blended(font, st, font_col );
   
    int w=cw[c]=text_surface->w;
    int h=globalh=text_surface->h;
    
    byte *b=(byte*)text_surface->pixels;
    
    //printf("char '%c', w=%d, h=%d\n",c,w,h);
    
    if (posx+w >=TSIZEX) {
      posx=2;
      posy+=CSIZEH+1;
      if ((posy+h)>=TSIZEY) { toobig=true; break;} // font too big!
    }
    
    cx[c]=posx; ct[c]=posy;
    cy[c]=-1;
    // copy char in texture
    for (int y=posy,j=0; y<posy+h; y++) 
    for (int x=posx; x<posx+w; x++, j+=4) 
    {
      byte alpha=b[j+3];
      if (alpha) {
        if (cy[c]==-1) cy[c]=y-1;
        ch[c]=y-cy[c]+2;
      }
      buffer[(y*TSIZEX+x)*4+0]=255;//b[j+0];
      buffer[(y*TSIZEX+x)*4+1]=255;//b[j+1];
      buffer[(y*TSIZEX+x)*4+2]=255;//b[j+2];
      buffer[(y*TSIZEX+x)*4+3]=alpha;//*8/10+50;
      // checkboard (debugging):
      if (DEBUG_MAPEDIT) buffer[(y*TSIZEX+x)*4+3]=((y<h/2+posy)==(x<w/2+posx))?100:57;

    }
    //if (ch[c]==0) ch[c]=1;
    posx+=w+1;
    if (text_surface) SDL_FreeSurface(text_surface);
    
  }
  if (!toobig) return true;
  CSIZEW--;
}
};
  
float Texture::u(char c, int zeroone){
  return (cx[c]+zeroone*(cw[c]))/double(TSIZEX);
};
float Texture::v(char c, int zeroone){
  return (cy[c]+zeroone*(ch[c]))/double(TSIZEY);
};

float Texture::multx(char c){
  if (!present[c]) return 1.0;
  return cw[c]/float(globalh);
}

float Texture::multy(char c, int zeroone){
  if (!present[c]) return 1.0;
  float centery = ct[c] + 0.5*globalh;
  
  if (zeroone) return ( cy[c] + ch[c] - centery   ) / (0.5*globalh);
  else return ( centery - cy[c]  ) / (0.5*globalh) ;

}

